classdef VEMalgo < handle
%==========================================================================
% VEM algorithm for multichannel audio source separation.
%
% S. Leglaive, R. Badeau, G. Richard. "Semi-blind Student's t source
% separation for multichannel audio convolutive mixtures", submitted for
% publication in Proc. of the European Signal Processing Conference (Eusipco),
% Kos Island, Greece, 2017.
%
%==========================================================================
% Copyright (c) 2017, Simon Leglaive, Télécom ParisTech, All Rights Reserved.
%
% This code uses the BSD License:
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
% * Redistributions of source code must retain the above copyright
% notice, this list of conditions and the following disclaimer.
% * Redistributions in binary form must reproduce the above copyright
% notice, this list of conditions and the following disclaimer in the
% documentation and/or other materials provided with the distribution.
% * Neither the name of Telecom ParisTech nor the names of its contributors
% may be used to endorse or promote products derived from this software without
% specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
% EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
% DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
% DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
% ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
% SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%==========================================================================

    properties
        x; % time domaine observations [T x I]

        % Model parameters
        a; % mixing filters [I x J x La]
        sigmaNoise; % additive noise variances vector [1 x I]
        lambda2; % student's t scale parameters [F x N x J]
        alpha; % student's t shape parameters [1 x J]
        W; % {J x 1} cell array containing the matrices of NMF bases for each source
        H; % {J x 1} cell array containing the matrices of NMF activations for each source

        % Variational distribution parameters
        m_post; % s_jfn mean under variational approximation [F x N x J]
        gamma_post; % s_jfn variance under variational approximation [F x N x J]
        delta; % v_jfn shape parameters under variational approximation [1 x J]
        beta; % v_jfn scale parameters under variational approximation [F x N x J]

        % Others
        wlen; % MDCT synthesis window length
        expectError; % expectation of the squared error between the model and the observations (sum_t e_i(t)) [1 x I]
        varFreeEnergy; % objective
        psiCorr; % troncated autocorrelation of time-frequency MDCT atoms
        G; % [F x (La + wlen - 1) x I x J] : contains the convolution of the MDCT atom at frequency f with the
           % mixing filters a_ij
        s; % [Ls x J] time domain sources
        sim; % [T x J x I] time domaine source images
    end

    methods

        function this = VEMalgo(varargin)
            this.Set(varargin{:});
        end

        function Set(this,varargin)
            st_input = struct(varargin{:});

            if isfield(st_input, 'x')
                this.x = st_input.x;
            end

            if isfield(st_input, 'a')
                this.a = st_input.a;
            end

            if isfield(st_input, 'sigmaNoise')
                this.sigmaNoise = st_input.sigmaNoise;
            end

            if isfield(st_input, 'alpha')
                this.alpha = st_input.alpha;
            end

            if isfield(st_input, 'psi')
                this.H = st_input.psi;
            end

            if isfield(st_input, 'wlen')
                this.wlen = st_input.wlen;
            end

            if isfield(st_input, 'W')
                this.W = st_input.W;
            end

            if isfield(st_input, 'H')
                this.H = st_input.H;
            end

            J = length(this.alpha);
            hop = this.wlen/2;

            [T, I] = size(this.x);
            La = size(this.a,3);
            Ls = T - La + 1;

            F = this.wlen/2;
            N = round(Ls/hop) - 1;

            if isfield(st_input, 'lambda')
                this.lambda2 = ones(F, N, J);
                for j=1:J
                    this.lambda2(:,:,j) = this.lambda2(:,:,j).*(st_input.lambda(j)^2);
                end
            else
                this.lambda2 = ones(F, N, J);
            end

            if isfield(st_input, 'W') && isfield(st_input, 'H')
                this.lambda2 = ones(F, N, J);
                for j=1:J
                    this.lambda2(:,:,j) = this.W{j}*this.H{j};
                end
            end

            this.m_post = zeros(F,N,J);
            this.gamma_post = zeros(F,N,J);

            this.delta = (this.alpha + 1)/2;
            this.beta = ones(F,N,J);

            this.s = zeros(Ls, J);
            this.sim = zeros(T, J, I);
            this.ComputePsiCorr;
            this.G = zeros(F, La + this.wlen - 1, I, J);
            this.ComputeGMat;
            this.expectError = zeros(1,I);

        end

        function [] = ComputePsiCorr(this)

            F = size(this.m_post,1);
            [T, ~] = size(this.x);
            La = size(this.a,3);
            Ls = T - La + 1;
            psiCorrTmp = zeros(F,max(this.wlen,La));
            for f=1:F
                psi_fn = this.GetMDCTAtom(f-1,0,Ls,this.wlen);
                tmp = xcorr(psi_fn(1:max(this.wlen,La)));
                psiCorrTmp(f,:) = tmp(max(this.wlen,La):end).';
            end
            this.psiCorr = psiCorrTmp;

        end

        function [] = EStepPCG(this)

            [~, I] = size(this.x);
            F = size(this.m_post,1);
            N = size(this.m_post,2);
            J = size(this.m_post,3);
            La = size(this.a,3);
            hop = this.wlen/2;

            %%%%%%%%%%%% E-step for v_jfn %%%%%%%%%%%%
            this.delta = (this.alpha + 1)/2;

            this.beta = (this.m_post.^2 + this.gamma_post)/2 + ...
                bsxfun(@times, this.lambda2, permute(this.alpha/2, [1,3,2]));

            %%%%%%%%%%%% E-step for s_jfn %%%%%%%%%%%%

            inv_v_post = bsxfun(@times, 1./(this.beta+eps), permute(this.delta, [1,3,2]));

            % compute gamma_post
            tmp = squeeze(sum(sum(this.G.^2,2)./permute(repmat(this.sigmaNoise', [1, 1, J, F]), [4,2,1,3]),3)); % F x J
            tmp = repmat(tmp, [1,1,N]);
            tmp = permute(tmp, [1,3,2]);
            this.gamma_post = 1./(inv_v_post + tmp);

            invD = diag(sparse(this.gamma_post(:)));

            % compute gradient
            epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
            frameInd = zeros(this.wlen+La-1,N); % column n of frameInd contains the indices of the signal to be considered for frame n
            for n=1:N
                frameInd(:,n) = (n-1)*hop + (1:this.wlen+La-1)';
            end
            epsilonMat = zeros(this.wlen+La-1,N,J,I);
            for i=1:I
                epsilon_i = epsilon(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                epsilonMat(:,:,:,i) = repmat(epsilon_i(frameInd), [1,1,J]); % (wlen + La - 1) x N x J x I
            end

            this.G = permute(this.G, [1,2,4,3]);
            grad = this.m_post.*inv_v_post - sum(mtimesx(this.G,epsilonMat),4);
            this.G = permute(this.G, [1,2,4,3]);

            % compute descent direction
            w = invD*grad(:);
            w = reshape(w, F, N, J);

            niter = 10;
            for iter = 1:niter

                % compute gamma
                w_mdct = this.ImdctMulti(w); % Ls x J
                w_filt = this.SrcImage(w_mdct); % T x J x I
                w_filt = squeeze(sum(w_filt, 2)); % T x I
                w_filtMat = zeros(this.wlen+La-1,N,J,I); % (wlen + La - 1) x N x J x I
                for i=1:I
                    w_filt_i = w_filt(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                    w_filtMat(:,:,:,i) = repmat(w_filt_i(frameInd), [1,1,J]);
                end

                this.G = permute(this.G, [1,2,4,3]);
                gamma = w.*inv_v_post + sum(mtimesx(this.G,w_filtMat),4);
                this.G = permute(this.G, [1,2,4,3]);

                % compute mu
                mu = real((w(:)'*grad(:))/(w(:)'*gamma(:)));

                % update source MDCT coefficients
                this.m_post = this.m_post - mu*w;

                % update source images
                this.ImdctMultiProp(1:J);
                this.SrcImageProp(1:J,(1:I));

                % compute gradient
                epsilon = this.x - sum(permute(this.sim,[1,3,2]),3); % T x I
                epsilonMat = zeros(this.wlen+La-1,N,J,I);
                for i=1:I
                    epsilon_i = epsilon(:,i)/this.sigmaNoise(i); % (wlen + La - 1) x N
                    epsilonMat(:,:,:,i) = repmat(epsilon_i(frameInd), [1,1,J]); % (wlen + La - 1) x N x J
                end

                this.G = permute(this.G, [1,2,4,3]);
                grad = this.m_post.*inv_v_post - sum(mtimesx(this.G,epsilonMat),4);
                this.G = permute(this.G, [1,2,4,3]);

                % preconditionning
                gradP = invD*grad(:);
                gradP = reshape(gradP, F, N, J);

                % compute alpha
                alpha_pcg = -real((gamma(:)'*gradP(:))/(w(:)'*gamma(:)));

                % compute w
                w = gradP + alpha_pcg*w;
            end
        end

        function [] = UpdateNoiseVar(this)
            [T,I] = size(this.x);
            this.sigmaNoise = 1/T*this.expectError;
        end

        function [] = UpdateLambda(this, str_model)

            if nargin < 2
                str_model = 'NMF';
            end

            J = size(this.m_post,3);

            if strcmp(str_model, 'sparse')
                F = size(this.m_post,1);
                N = size(this.m_post,2);
                tmp = F*N ./ this.delta ./ sum(sum(permute(1./(this.beta + eps), [3 1 2]),3),2).';
                for j=1:J
                    this.lambda2(:,:,j) = tmp(j);
                end
            elseif strcmp(str_model, 'unconstrained')
                this.lambda2 = bsxfun(@times, this.beta, permute(1./this.delta, [1,3,2]));
            elseif strcmp(str_model, 'NMF')
                for j=1:J
                    E = this.UpdateNMF(j, 50);
                    this.lambda2(:,:,j) = this.W{j}*this.H{j};
                end
            else
                error('Unknown source model');
            end

        end

        function [E] = UpdateNMF(this, j, iter_num)

            if nargin < 3 || isempty(iter_num)
                iter_num = 50;
            end

            [F, N] = size(this.beta(:,:,j));

            E = zeros(iter_num,1);
            WH = this.W{j}*this.H{j} + eps;

            for iter = 1:iter_num
                % update of W
                this.W{j} = this.W{j}.* ( (1./WH * this.H{j}' + eps ) ./ ...
                    ( ( this.delta(j)./this.beta(:,:,j)) * this.H{j}' + eps) );

                % normalization
                w = sum(this.W{j},1);
                d = diag(1./w);
                this.W{j} = this.W{j}*d; % Normalizatiosn of columns of W
                d = diag(w);
                this.H{j} = d*this.H{j}; % Energy transfer to H

                WH = this.W{j}*this.H{j} + eps;

                % update of H
                this.H{j} = this.H{j}.* ( ( this.W{j}' * (1./WH) + eps ) ./ ...
                    ( this.W{j}' * ( this.delta(j)./this.beta(:,:,j) ) + eps) );

                WH = this.W{j}*this.H{j} + eps;

                % Error with IS divergence:
                E(iter) = sum(sum( WH ./ (this.beta(:,:,j)/this.delta(j) + eps) - log( WH + eps ) )); % normalized error

            end

        end

        function [] = ComputeExpectError(this)

            epsilon = this.x - squeeze(sum(this.sim,2)); % T x I
            
            normG = permute(sum(this.G.^2,2), [1,2,4,3]); % F x 1 x J x I
            gamma_normG = bsxfun(@times, this.gamma_post, normG); % F x N x J x I
            
            this.expectError = sum(epsilon.^2,1) + permute(sum(sum(sum(gamma_normG,1),2),3), [3,4,1,2]);
        end

        function [] = ComputeVarFreeEnergy(this, b_computeExpectError)

            if nargin < 2
                b_computeExpectError = 1;
            end

            [T,~] = size(this.x);
            F = size(this.m_post,1);
            N = size(this.m_post,2);

            if b_computeExpectError
                this.ComputeExpectError;
            end

             this.varFreeEnergy = -0.5*sum(this.expectError./this.sigmaNoise + T*log(this.sigmaNoise))...
                - 0.5*sum(sum(sum( log(this.beta./(this.gamma_post)) ...
                    + bsxfun(@times, (this.m_post.^2 + this.gamma_post)./(this.beta), permute(this.delta, [1,3,2]) ) ))) ...
                - sum(sum(sum( bsxfun(@times, log(this.beta./(this.lambda2)), permute(this.alpha/2, [1,3,2])) ...
                    + bsxfun(@times, this.lambda2./(this.beta), permute(this.alpha/2.*this.delta, [1,3,2])) )));
        end

        function [ psi ] = GetMDCTAtom(this, f, n, siglen, wlen )

            F = wlen/2;
            win = sin((.5:wlen-.5)/wlen*pi); % sine window
            hop = wlen/2; % hop size = wlen/2;

            frameInd = (n*hop+1 : n*hop + wlen)';
            if frameInd(end)>siglen
                error('out of signal')
            end
            psi = zeros(siglen,1);
            psi(frameInd) = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:wlen-1) + 0.5 + wlen/4) * 2*pi/wlen ) );

        end

        function [] = ComputeGMat(this)

            F = size(this.m_post,1);
            J = size(this.m_post,3);
            [~,I] = size(this.x);
            win = sin((.5:this.wlen-.5)/this.wlen*pi); % sine window

            for f=0:F-1
                for j=1:J
                    psi_f = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:this.wlen-1) + 0.5 + this.wlen/4) * 2*pi/this.wlen ) ); % wlen x 1
                    a_sq = permute(this.a(:,j,:),[3,1,2]); % La x I
                    this.G(f+1,:,:,j) = this.FFTConv(psi_f'*ones(1,I),a_sq); % (La + wlen - 1) x I
                end
            end
        end

        function [] = ImdctMultiProp(this, srcInd)

            [F, N, J] = size(this.m_post);
            Ls = size(this.s,1);

            if nargin<2
                srcInd = 1:J;
            end

            hop = this.wlen/2; % hop size = wlen/2;
            win = sin((.5:this.wlen-.5)/this.wlen*pi); % sine window

            X_ext = [this.m_post(:,:,:) ; -this.m_post(end:-1:1,:,:)];
            x_rec = zeros(hop*(N+1), 1);

            for j = srcInd
                x_rec(:) = 0;
                frames = real(ifft(X_ext(:,:,j).*(exp(1i*2*pi*(0.5+this.wlen/4)*(0:this.wlen-1)'/this.wlen)*ones(1,N)))...
                    .*(exp(1i*pi*(0.5+this.wlen/4+(0:this.wlen-1)')/this.wlen)*ones(1,N)))*sqrt(this.wlen);
                for n = 1:N
                    frameInd = (n-1)*hop + (1:this.wlen);
                    x_rec(frameInd) = x_rec(frameInd) + win'.*frames(:,n);
                end
                this.s(:,j) = x_rec(1:Ls,:,:);
            end
        end

        function [x] = ImdctMulti(this, X)

            [F, N, J] = size(X);
            Ls = size(this.s,1);
            x = zeros(Ls,J);

            hop = this.wlen/2; % hop size = wlen/2;
            win = sin((.5:this.wlen-.5)/this.wlen*pi); % sine window

            X_ext = [X(:,:,:) ; -X(end:-1:1,:,:)];
            x_rec = zeros(hop*(N+1), 1);

            for j = 1:J
                x_rec(:) = 0;
                frames = real(ifft(X_ext(:,:,j).*(exp(1i*2*pi*(0.5+this.wlen/4)*(0:this.wlen-1)'/this.wlen)*ones(1,N)))...
                    .*(exp(1i*pi*(0.5+this.wlen/4+(0:this.wlen-1)')/this.wlen)*ones(1,N)))*sqrt(this.wlen);
                for n = 1:N
                    frameInd = (n-1)*hop + (1:this.wlen);
                    x_rec(frameInd) = x_rec(frameInd) + win'.*frames(:,n);
                end
                x(:,j) = x_rec(1:Ls,:,:);
            end
        end

        function [] = SrcImageProp(this, srcInd, micInd)
            % s : [Ls x J]
            % si : [T x J x I]

            J = length(srcInd);
            I = length(micInd);
            T = size(this.x,1);
            Ls = size(this.s,1);
            La = size(this.a,3);

            s = this.s(:, srcInd);
            a = this.a(micInd, srcInd, :);

            s_resh = zeros(Ls,J,I);
            for i=1:I
                s_resh(:,:,i) = s;
            end
            s_resh = reshape(s_resh, Ls, I*J);
            a_resh = reshape(permute(a,[3,2,1]), La, I*J);
            si = this.FFTConv(s_resh,a_resh);
            this.sim(:,srcInd, micInd) = reshape(si, T, J, I);

        end

        function [xi] = SrcImage(this, x)
            % s : [Ls x J]
            % si : [T x J x I]
            J = size(x,2);
            [T, I] = size(this.x);
            Ls = size(this.s,1);
            La = size(this.a,3);

            x_resh = zeros(Ls,J,I);
            for i=1:I
                x_resh(:,:,i) = x;
            end
            x_resh = reshape(x_resh, Ls, I*J);
            a_resh = reshape(permute(this.a,[3,2,1]), La, I*J);
            xi = this.FFTConv(x_resh,a_resh);
            xi = reshape(xi, T, J, I);

        end

        function y = FFTConv(this, x, h)
            T = length(x) + length(h) - 1;
            Nfft = 2^nextpow2(T);
            y = real(ifft(fft(x, Nfft).*fft(h, Nfft), Nfft));
            y = y(1:T,:);
        end

    end
end
