function [ X ] = mdct_multi( x, wlen, b_fft )

% MDCT_MULTI Multichannel Modified Discrete Cosine Transform (MDCT) using
% half-overlapping sine windows.
%
% X=mdct_multi(x)
% X=mdct_multi(x, wlen)
% X=mdct_multi(x, wlen, b_fft)
%
% Inputs:
% x: [siglen x I] matrix containing the I channels time-domain mixture signals
% with siglen samples.
% wlen: window length (default: 1024 samples or 64 ms at 16 kHz.
% b_fft: if 1, fft is used to compute MDCT.
%
% Output:
% X: [F x N x I] matrix containing the MDCT coefficients with F
% frequency bins and N time frames.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simon Leglaive, Tlcom ParisTech, 2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin<1
    error('Not enough input arguments.');
end
if nargin<2
    wlen = 1024; % window length
    b_fft = 0;
end
if nargin<3
    b_fft = 0;
end

[siglen,I] = size(x);

if I>siglen
    error('The signals must be within columns.');
end

if wlen~=2*floor(wlen/2)
    error('The window length must be a multiple of 2.');
end

F = wlen/2; % number of frequency bins = wlen/2
hop = wlen/2; % hop size = wlen/2;

win = sin((.5:wlen-.5)/wlen*pi); % sine window

x = zeroPad(x, wlen);

N = size(x,1)/hop - 1; % number of frames

frameInd = zeros(wlen,N); % column n of frameInd contains the indices of the signal to be considered for frame n
for n=1:N
    frameInd(:,n) = ((n-1)*hop+1 : (n-1)*hop + wlen)';
end

X = zeros(F,N,I);

if ~b_fft
    
    % [F x wlen] matrix of MDCT transform for a frame
    T = zeros(F,wlen);
    for f = 0:F-1,
        T(f+1,:) = win.*( sqrt(2/F) * cos( (f+1/2) * ((0:wlen-1) + 0.5 + wlen/4) * 2*pi/wlen ) );
    end
    
    % compute MDCT
    for i = 1:I
        tmp = x(:,i);
        X(:,:,i) = T*tmp(frameInd);
    end
    
else
    
    % compute MDCT
    for i = 1:I
        x_i = x(:,i);
        frameMat = x_i(frameInd);
        frameMat = frameMat.*repmat(win'.*(exp(-1i*pi*(0:wlen-1)'/wlen)),1,N);
        framefftMat = fft(frameMat);
        X(:,:,i) = sqrt(2/F)*real(framefftMat(1:F,:).*repmat(exp(-1i*2*pi/wlen*(0.5+wlen/4)*(0.5:F-0.5)'),1,N));
    end
    
end

end

